# SPDX-License-Identifier: GPL-3.0-or-later
"""
Custom canvas implementation.

Acts as an interface between matplotlib and GObject and contains a custom
implementation of a `SpanSelector` as well as a dummy toolbar used for
interactive navigation in conjunction with graphs-specific structures.
"""
import math

from gi.repository import Adw, Gdk, Gio, Graphs, Gtk

from graphs.figure import Figure

from matplotlib import RcParams, backend_tools as tools
from matplotlib.backend_bases import (
    FigureCanvasBase,
    MouseEvent,
    NavigationToolbar2,
)
from matplotlib.backends.backend_gtk4cairo import FigureCanvas
from matplotlib.widgets import SpanSelector

_SCROLL_SCALE = 1.08


class Canvas(Graphs.Canvas, FigureCanvas):
    """Custom Canvas."""

    __gtype_name__ = "GraphsPythonCanvas"

    def __init__(
        self,
        style_params: tuple[RcParams, dict],
        items: Gio.ListModel,
        interactive: bool = True,
        figure_settings: Graphs.FigureSettings = None,
    ):
        """
        Create the canvas.

        Create figure, axes and define rubberband_colors based on the current
        style context. Bind `items` to `data.items` and all figure settings
        attributes to their respective values.
        """
        Graphs.Canvas.__init__(
            self,
            hexpand=True,
            vexpand=True,
        )
        self._idle_draw_id = 0
        self.set_draw_func(self._draw_func)
        FigureCanvasBase.__init__(
            self,
            figure=Figure(style_params, items, self, figure_settings),
        )
        self._rubberband_rect = None

        self.connect("notify::scale-factor", self._update_device_pixel_ratio)
        self.connect("resize", self._on_resize)

        # Handle stuff only used if the canvas is interactive
        if interactive:
            self._setup_interactive()

    def _setup_interactive(self):
        self._ctrl_held, self._shift_held = False, False
        self._xfrac, self._yfrac = None, None
        self.mpl_connect("pick_event", self._on_pick)
        self.mpl_connect("motion_notify_event", self._set_mouse_fraction)
        self._make_ticklabels_pickable()

        # Reference is created by the toolbar itself
        _DummyToolbar(self)

        click = Gtk.GestureClick()
        click.set_button(0)  # All buttons.
        click.connect("pressed", self.button_press_event)
        click.connect("update", self.handle_touch_update)
        click.connect("released", self.button_release_event)
        self.add_controller(click)

        motion = Gtk.EventControllerMotion()
        motion.connect("motion", self.motion_notify_event)
        motion.connect("enter", self.enter_notify_event)
        motion.connect("leave", self.leave_notify_event)
        self.add_controller(motion)

        scroll = Gtk.EventControllerScroll.new(
            Gtk.EventControllerScrollFlags.BOTH_AXES,
        )
        scroll.connect("scroll", self.scroll_event)
        scroll.connect("scroll-end", self.toolbar.push_current)
        self.add_controller(scroll)

        zoom = Gtk.GestureZoom.new()
        zoom.connect("scale-changed", self.zoom_event)
        zoom.connect("end", self.end_zoom_event)
        self.add_controller(zoom)

        def rgba_to_tuple(rgba):
            return (rgba.red, rgba.green, rgba.blue, rgba.alpha)

        style_manager = Adw.StyleManager()
        rgba = style_manager.get_accent_color_rgba()
        self.rubberband_edge_color = rgba_to_tuple(rgba)
        rgba.alpha = 0.2
        self.rubberband_fill_color = rgba_to_tuple(rgba)
        self.highlight = _Highlight(self)
        self.connect("notify::mode", self._on_mode_change)
        self._on_mode_change()

        for item in ("min", "max"):
            self.connect(
                f"notify::{item}-selected",
                lambda _a,
                _b: self.highlight.load(self),
            )

    def _make_ticklabels_pickable(self) -> None:
        """Make all tick labels pickable."""
        for ax in self.figure.axes:
            for label in (ax.get_xticklabels() + ax.get_yticklabels()):
                label.set_picker(True)

    def handle_touch_update(self, controller: Gtk.GestureClick, _data) -> None:
        """
        Handle an update event for GtkGestureClick motion.

        This is needed for touch screen devices to handle gestures properly.
        """
        if not controller.get_point()[0]:  # If touch event
            coords = controller.get_bounding_box_center()
            x, y = coords.x, coords.y
            MouseEvent(
                "motion_notify_event",
                self,
                *self._mpl_coords((x, y)),
            )._process()

    def key_press_event(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> None:
        """Handle key press event."""
        if keyval in {65507, 65508}:  # Control_L or Control_R
            self._ctrl_held = True
        elif keyval in {65505, 65506}:  # Left or right Shift
            self._shift_held = True
        else:  # Prevent keys from being true with key combos
            self._ctrl_held = False
        super().key_press_event(controller, keyval, keycode, state)

    def key_release_event(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> None:
        """Handle key release event."""
        self._ctrl_held = False
        self._shift_held = False
        super().key_release_event(controller, keyval, keycode, state)

    def scroll_event(
        self,
        controller: Gtk.EventControllerScroll,
        dx: float,
        dy: float,
    ) -> None:
        """
        Handle scroll event.

        Updates only axes with independent coordinate systems to prevent uneven
        scaling:
        - X-limits: _axis and _top_left_axis (independent x ax.)
        - Y-limits: _axis, _right_axis, and _top_right_axis (independent y ax.)
        """
        if self._ctrl_held:
            self.zoom(1 / _SCROLL_SCALE if dy > 0 else _SCROLL_SCALE)
        else:
            if self._shift_held:
                dx, dy = dy, dx
            if controller.get_unit() == Gdk.ScrollUnit.WHEEL:
                dx *= 10
                dy *= 10

            for ax in [self.figure.axis, self.figure.top_left_axis]:
                xmin, xmax = ax.get_xlim()
                scale = Graphs.scale_from_string(ax.get_xscale())
                xmin, xmax = self._calculate_pan_values(xmin, xmax, scale, dx)
                ax.set_xlim(xmin, xmax)

            for ax in [
                self.figure.axis,
                self.figure.right_axis,
                self.figure.top_right_axis,
            ]:
                ymin, ymax = ax.get_ylim()
                scale = Graphs.scale_from_string(ax.get_yscale())
                ymin, ymax = self._calculate_pan_values(ymin, ymax, scale, -dy)
                ax.set_ylim(ymin, ymax)

        self.toolbar.push_current()
        super().scroll_event(controller, dx, dy)

    def zoom_event(
        self,
        controller: Gtk.GestureZoom,
        scale: float,
    ) -> None:
        """Handle zoom event."""
        coords = controller.get_bounding_box_center()
        x, y = coords.x, coords.y
        event = MouseEvent(
            "motion_notify_event",
            self,
            *self._mpl_coords((x, y)),
        )
        self._set_mouse_fraction(event)
        scale = 1 + 0.015 * (scale - 1)
        if scale > 5 or scale < 0.2:
            # Don't scale if ridiculous values are registered
            return
        self.zoom(scale)

    def end_zoom_event(self, controller: Gtk.GestureZoom, _sequence) -> None:
        """
        End the zoom event.

        Pushes the canvas to the stack, and emits a `release` signal cancel out
        registered touches from touchscreen devices.
        """
        coords = controller.get_bounding_box_center()
        x, y = coords.x, coords.y
        MouseEvent(
            "button_release_event",
            self,
            *self._mpl_coords((x, y)),
            1,
        )._process()
        self.toolbar.push_current()

    def enter_notify_event(
        self,
        controller: Gtk.EventControllerMotion,
        x: float,
        y: float,
    ) -> None:
        """Process pointer entry."""
        self.grab_focus()
        super().enter_notify_event(controller, x, y)

    def _set_mouse_fraction(self, event) -> None:
        """Set the mouse coordinate in terms of fraction of the canvas."""
        if event.inaxes is not None:
            xlim = self.figure.top_right_axis.get_xlim()
            ylim = self.figure.top_right_axis.get_ylim()
            self._xfrac = Graphs.get_fraction_at_value(
                event.xdata,
                xlim[0],
                xlim[1],
                self.figure.props.top_scale,
            )
            self._yfrac = Graphs.get_fraction_at_value(
                event.ydata,
                ylim[0],
                ylim[1],
                self.figure.props.right_scale,
            )
        else:
            self._xfrac, self._yfrac = None, None

    def zoom(self, scaling: float = 1.25, respect_mouse: bool = True) -> None:
        """
        Zoom with given scaling.

        Update all axes' limits in respect to the current mouse position,
        updates only axes with independent coordinate systems to prevent uneven
        scaling:
        - X-limits: _axis and _top_left_axis (independent x-ax.)
        - Y-limits: _axis, _right_axis, and _top_right_axis (independent y-ax.)
        """
        if not respect_mouse:
            self._xfrac, self._yfrac = 0.5, 0.5
        if self._xfrac is None or self._yfrac is None:
            return

        for ax in [self.figure.axis, self.figure.top_left_axis]:
            ax.set_xlim(
                self._calculate_zoomed_values(
                    self._xfrac,
                    Graphs.scale_from_string(ax.get_xscale()),
                    ax.get_xlim(),
                    scaling,
                ),
            )
        for ax in [
            self.figure.axis,
            self.figure.right_axis,
            self.figure.top_right_axis,
        ]:
            ax.set_ylim(
                self._calculate_zoomed_values(
                    self._yfrac,
                    Graphs.scale_from_string(ax.get_yscale()),
                    ax.get_ylim(),
                    scaling,
                ),
            )

        self.queue_draw()

    @staticmethod
    def _calculate_pan_values(
        current_min: float,
        current_max: float,
        scale: Graphs.Scale,
        panspeed: float,
    ) -> tuple[float, float]:
        """Calculate axis values required for panning."""
        pan_scale = 0.003

        value1 = Graphs.get_value_at_fraction(
            panspeed * pan_scale,
            current_min,
            current_max,
            scale,
        )
        value2 = Graphs.get_value_at_fraction(
            1 + panspeed * pan_scale,
            current_min,
            current_max,
            scale,
        )

        if scale == Graphs.Scale.INVERSE:
            value1, value2 = value2, value1

        return value1, value2

    @staticmethod
    def _calculate_zoomed_values(
        fraction: float,
        scale: Graphs.Scale,
        limit: float,
        zoom_factor: float,
    ) -> tuple[float, float]:
        """
        Calculate zoomed values.

        Calculates the coordinates of the canvas after a zoom gesture
        has  been ezoomed.
        """
        min_, max_ = limit[0], limit[1]
        value1 = Graphs.get_value_at_fraction(
            fraction - fraction / zoom_factor,
            min_,
            max_,
            scale,
        )
        value2 = Graphs.get_value_at_fraction(
            fraction + (1 - fraction) / zoom_factor,
            min_,
            max_,
            scale,
        )
        if scale == Graphs.Scale.INVERSE:
            value1, value2 = value2, value1
        return value1, value2

    def _on_resize(self, widget, width, height) -> None:
        """Handle the legend update on window resizes."""
        self.resize_event(widget, width, height)
        self.figure.update_legend()

    def _on_pick(self, event) -> None:
        """Emit edit-request signal for picked label, tick or title."""
        artist = event.artist

        if not hasattr(artist, "id"):
            artist.id = self._determine_figure_setting(artist)

        self.emit("edit_request", artist.id)

    def _determine_figure_setting(self, artist) -> str:
        """
        Determine the figure settings to be used on the tick after pick event.

        Determines the artist's position to generate an artist id that matches
        with the appropriate limits-widget figure_settings.

        The artist position is a tuple where one element is an integer (0 or 1)
        indicating the axis side, and the other is a numpy float indicating the
        value of the clicked tick.
        The position of the integer determines the axis:
        - Integer at index 0: X-axis (0=left, 1=right)
        - Integer at index 1: Y-axis (0=bottom, 1=top)
        """
        artist_position = artist.get_position()

        # X axis
        if isinstance(artist_position[0], int):
            position, label_value = artist_position
            side = "left" if position == 0 else "right"

        # Y-axis
        elif isinstance(artist_position[1], int):
            label_value, position = artist_position
            side = "bottom" if position == 0 else "top"

        min_val = self.figure.get_property(f"min_{side}")
        max_val = self.figure.get_property(f"max_{side}")
        midpoint = (min_val + max_val) / 2
        position_type = "max" if label_value > midpoint else "min"

        return f"{position_type}_{side}"

    # Overwritten function - do not change name
    def _post_draw(self, _widget, context) -> None:
        """Allow custom rendering extensions."""
        if self._rubberband_rect is not None:
            self._draw_rubberband(context)

    def _draw_rubberband(self, ctx) -> None:
        """
        Implement custom rubberband.

        Draw a rubberband matching libadwaitas style, where `_rubberband_rect`
        is set.
        """
        radius = 6
        x0, y0, width, height = (
            dim / self.device_pixel_ratio for dim in self._rubberband_rect
        )
        x1 = x0 + width
        y1 = y0 + height

        if x1 < x0:
            x0, x1 = x1, x0
        if y1 < y0:
            y0, y1 = y1, y0

        degrees = math.pi / 180

        ctx.new_sub_path()
        ctx.arc(x1 - radius, y0 + radius, radius, -90 * degrees, 0 * degrees)
        ctx.arc(x1 - radius, y1 - radius, radius, 0 * degrees, 90 * degrees)
        ctx.arc(x0 + radius, y1 - radius, radius, 90 * degrees, 180 * degrees)
        ctx.arc(x0 + radius, y0 + radius, radius, 180 * degrees, 270 * degrees)
        ctx.close_path()

        ctx.set_line_width(1)
        ctx.set_source_rgba(*self.rubberband_fill_color)
        ctx.fill_preserve()
        ctx.set_source_rgba(*self.rubberband_edge_color)
        ctx.stroke()

    def _on_mode_change(self, *_args) -> None:
        highlight_enabled = self.props.mode == Graphs.Mode.SELECT
        self.highlight.set_active(highlight_enabled)
        self.highlight.set_visible(highlight_enabled)
        self.queue_draw()


class _DummyToolbar(NavigationToolbar2):
    """Custom Toolbar implementation."""

    # Overwritten function - do not change name
    def _zoom_pan_handler(self, event) -> None:
        mode = self.canvas.props.mode
        if event.button == 2:
            event.button = 1
            mode = Graphs.Mode.PAN
        elif event.button != 1:
            return
        if mode == Graphs.Mode.PAN:
            if event.name == "button_press_event":
                self.press_pan(event)
            elif event.name == "button_release_event":
                self.release_pan(event)
        elif mode == Graphs.Mode.ZOOM:
            if event.name == "button_press_event":
                self.press_zoom(event)
            elif event.name == "button_release_event":
                self.release_zoom(event)

    # Overwritten function - do not change name
    def _update_cursor(self, event) -> None:
        mode = self.canvas.props.mode
        if event.inaxes and event.inaxes.get_navigate():
            if mode == Graphs.Mode.PAN and \
                    self._last_cursor != tools.Cursors.MOVE:
                self.canvas.set_cursor(tools.Cursors.MOVE)
                self._last_cursor = tools.Cursors.MOVE
            elif mode == Graphs.Mode.ZOOM and \
                    self._last_cursor != tools.Cursors.SELECT_REGION:
                self.canvas.set_cursor(tools.Cursors.SELECT_REGION)
                self._last_cursor = tools.Cursors.SELECT_REGION
        elif self._last_cursor != tools.Cursors.POINTER:
            self.canvas.set_cursor(tools.Cursors.POINTER)
            self._last_cursor = tools.Cursors.POINTER

    # Overwritten function - do not change name
    def drag_pan(self, event) -> None:
        """Handle dragging in pan/zoom mode."""
        for ax in self._pan_info.axes:
            # Using the recorded button at the press is safer than the current
            # button, as multiple buttons can get pressed during motion.
            # Use custom drag_pan that maxes sure limits are set in right order
            # even on inverted scale
            self.ax_drag_pan(
                ax,
                self._pan_info.button,
                event.key,
                event.x,
                event.y,
            )
        self.canvas.draw_idle()

    @staticmethod
    def ax_drag_pan(self, button, key: str, x: float, y: float) -> None:
        """
        Handle mouse events during a pan operation.

        Notes
        -----
        This is intended to be overridden by new projection types.
        """
        points = self._get_pan_points(button, key, x, y)
        if points is not None:
            # Max and min needs to be defined at correct position for this to
            # work with inverted scaling
            ylim = points[:, 1]
            xlim = points[:, 0]
            self.set_xlim(min(xlim), max(xlim))
            self.set_ylim(min(ylim), max(ylim))

    # Overwritten function - do not change name
    def draw_rubberband(self, _event, x0, y0, x1, y1) -> None:
        self.canvas._rubberband_rect = [
            int(val) for val in
            (x0, self.canvas.figure.bbox.height - y0, x1 - x0, y0 - y1)
        ]
        self.canvas.queue_draw()

    # Overwritten function - do not change name
    def remove_rubberband(self) -> None:
        self.canvas._rubberband_rect = None
        self.canvas.queue_draw()

    # Overwritten function - do not change name
    def push_current(self, *_args) -> None:
        """Use custom functionality for the view clipboard."""
        self.canvas.highlight.load(self.canvas)
        for direction in ("bottom", "left", "top", "right"):
            self.canvas.figure.notify(f"min-{direction}")
            self.canvas.figure.notify(f"max-{direction}")
        self.canvas.emit("view_changed")

    # Overwritten function - do not change name
    def save_figure(self) -> None:
        pass


class _Highlight(SpanSelector):

    def __init__(self, canvas: Canvas):
        super().__init__(
            canvas.figure.top_right_axis,
            lambda _x,
            _y: self.apply(canvas),
            "horizontal",
            useblit=True,
            props={
                "facecolor": canvas.rubberband_fill_color,
                "edgecolor": canvas.rubberband_edge_color,
                "linewidth": 1,
            },
            handle_props={"linewidth": 0},
            interactive=True,
            drag_from_anywhere=True,
        )
        canvas.figure.top_right_axis.callbacks.connect(
            "xlim_changed",
            lambda _x: self.load(canvas),
        )
        self.load(canvas)

    def load(self, canvas: Canvas) -> None:
        xmin, xmax = canvas.figure.top_left_axis.get_xlim()
        scale = canvas.figure.props.top_scale
        self.extents = (
            Graphs.get_value_at_fraction(
                canvas.get_min_selected(),
                xmin,
                xmax,
                scale,
            ),
            Graphs.get_value_at_fraction(
                canvas.get_max_selected(),
                xmin,
                xmax,
                scale,
            ),
        )

    def apply(self, canvas: Canvas) -> None:
        xmin, xmax = canvas.figure.top_left_axis.get_xlim()
        low, high = self.extents
        self.extents = max(xmin, low), min(xmax, high)

        for prefix, value in zip(["min_", "max_"], self.extents):
            canvas.set_property(
                prefix + "selected",
                Graphs.get_fraction_at_value(
                    value,
                    xmin,
                    xmax,
                    canvas.figure.props.top_scale,
                ),
            )



# SPDX-License-Identifier: GPL-3.0-or-later
"""Main window."""
from gi.repository import Graphs

from graphs import misc
from graphs.canvas import Canvas
from graphs.data import Data

from matplotlib import rcParams, rcParamsDefault

CSS_TEMPLATE = """
.canvas-view#{name} {{
    background-color: {background_color};
    color: {color};
}}
"""
_SCALES = [f"{d}_scale" for d in misc.DIRECTIONS]
_LABELS = [f"{d}_label" for d in misc.DIRECTIONS]
_LIMITS = [lim.replace("-", "_") for lim in misc.LIMITS]
_FIGURE_PROPERTIES = [
    "title",
    "legend",
    "legend_position",
    "hide_unselected",
] + _SCALES + _LABELS + _LIMITS


class PythonWindow(Graphs.Window):
    """The main window class."""

    __gtype_name__ = "GraphsPythonWindow"

    def __init__(self):
        super().__init__(data=Data())
        self.props.data.connect(
            "style-changed",
            self._on_style_changed,
        )
        key_controller = self.props.key_controller
        key_controller.connect("key-pressed", self._on_key_press_event)
        key_controller.connect("key-released", self._on_key_release_event)

        zoom_in_action = self.lookup_action("zoom-in")
        zoom_in_action.connect("activate", self._on_zoom_in)
        zoom_out_action = self.lookup_action("zoom-out")
        zoom_out_action.connect("activate", self._on_zoom_out)

        self._reload_canvas()

    def _on_style_changed(
        self,
        data: Graphs.Data,
        recolor_items: bool,
    ) -> None:
        """Handle style change."""
        if not recolor_items:
            self._reload_canvas()
            return

        old_style = data.get_old_selected_style_params()
        new_style = data.get_selected_style_params()

        old_cycle = old_style[0]["axes.prop_cycle"].by_key()["color"]
        new_cycle = new_style[0]["axes.prop_cycle"].by_key()["color"]
        old_err_cycle = old_style[1]["errorbar.color_cycle"].by_key()["color"]
        new_err_cycle = new_style[1]["errorbar.color_cycle"].by_key()["color"]

        count = 0
        errbar_count = 0
        for item in data:
            item.reset(old_style, new_style)

            if not isinstance(item, (Graphs.DataItem, Graphs.EquationItem)) \
                    or item.get_color() not in old_cycle:
                continue

            count %= len(new_cycle)
            item.set_color(new_cycle[count])
            count += 1

            if not isinstance(item, Graphs.DataItem) \
                    or not item.has_xerr() and not item.has_yerr() \
                    or item.get_errcolor() not in old_err_cycle:
                continue

            errbar_count %= len(new_err_cycle)
            item.set_errcolor(new_err_cycle[errbar_count])
            errbar_count += 1

        self._reload_canvas()

    def _on_key_press_event(self, *args) -> None:
        """Handle key press event."""
        if self.props.canvas is not None:
            self.props.canvas.key_press_event(*args)

    def _on_key_release_event(self, *args) -> None:
        """Handle key release event."""
        if self.props.canvas is not None:
            self.props.canvas.key_release_event(*args)

    def _on_edit_request(self, _canvas, label_id: str) -> None:
        """Handle edit request."""
        self.open_figure_settings(label_id)

    def _on_view_changed(self, _canvas) -> None:
        """Handle view change."""
        self.props.data.add_view_history_state()

    def _on_zoom_in(self, _a, _b) -> None:
        if self.props.canvas is not None:
            self.props.canvas.zoom(1.15, False)

    def _on_zoom_out(self, _a, _b) -> None:
        if self.props.canvas is not None:
            self.props.canvas.zoom(1 / 1.15, False)

    def _reload_canvas(self) -> None:
        """Reload the canvas."""
        rcParams.update(rcParamsDefault)
        params = self.props.data.get_selected_style_params()

        figure_settings = self.props.data.get_figure_settings()
        canvas = Canvas(params, self.props.data, True, figure_settings)

        for prop in _FIGURE_PROPERTIES:
            figure_settings.bind_property(prop, canvas.figure, prop, 1)

        for prop in ("min_selected", "max_selected"):
            figure_settings.bind_property(prop, canvas, prop, 1 | 2)

        canvas.connect("edit-request", self._on_edit_request)
        canvas.connect("view-changed", self._on_view_changed)

        # Set headerbar color and contrast
        css = CSS_TEMPLATE.format(
            name=self.props.content_view.get_name(),
            background_color=params[0]["figure.facecolor"],
            color=params[0]["text.color"],
        )
        self.props.css_provider.load_from_string(css)

        self.set_canvas(canvas)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Style editor Box."""
import contextlib

from cycler import cycler

from gi.repository import Adw, Gio, Graphs, Gtk, Pango

from graphs import misc, style_io

STYLE_DICT = {
    "linestyle": ["lines.linestyle"],
    "linewidth": ["lines.linewidth"],
    "markers": ["lines.marker"],
    "markersize": ["lines.markersize"],
    "draw_frame": [
        "axes.spines.bottom",
        "axes.spines.left",
        "axes.spines.top",
        "axes.spines.right",
    ],
    "tick_direction": ["xtick.direction", "ytick.direction"],
    "minor_ticks": ["xtick.minor.visible", "ytick.minor.visible"],
    "major_tick_width": ["xtick.major.width", "ytick.major.width"],
    "minor_tick_width": ["xtick.minor.width", "ytick.minor.width"],
    "major_tick_length": ["xtick.major.size", "ytick.major.size"],
    "minor_tick_length": ["xtick.minor.size", "ytick.minor.size"],
    "tick_bottom": ["xtick.bottom"],
    "tick_left": ["ytick.left"],
    "tick_top": ["xtick.top"],
    "tick_right": ["ytick.right"],
    "tick_labels": ["ticklabels"],
    "show_grid": ["axes.grid"],
    "grid_linewidth": ["grid.linewidth"],
    "value_padding": [
        "xtick.major.pad",
        "xtick.minor.pad",
        "ytick.major.pad",
        "ytick.minor.pad",
    ],
    "label_padding": ["axes.labelpad"],
    "title_padding": ["axes.titlepad"],
    "axis_width": ["axes.linewidth"],
    "text_color": [
        "text.color",
        "axes.labelcolor",
        "xtick.labelcolor",
        "ytick.labelcolor",
    ],
    "tick_color": ["xtick.color", "ytick.color"],
    "axis_color": ["axes.edgecolor"],
    "grid_color": ["grid.color"],
    "grid_opacity": ["grid.alpha"],
    "background_color": ["axes.facecolor"],
    "outline_color": ["figure.facecolor", "figure.edgecolor"],
    "errorbar_capsize": ["errorbar.capsize"],
    "errorbar_capthick": ["errorbar.capthick"],
    "errorbar_linewidth": ["errorbar.linewidth"],
    "errorbar_barsabove": ["errorbar.barsabove"],
}
VALUE_DICT = {
    "linestyle": misc.LINESTYLES,
    "markers": misc.MARKERSTYLES,
    "tick_direction": ["in", "out"],
}
FONT_STYLE_DICT = {
    0: "normal",
    1: "oblique",
    2: "italic",
}
FONT_VARIANT_DICT = {
    0: "normal",
    1: "small-caps",
}


class StyleEditorBox(Graphs.StyleEditorBox):
    """Style editor widget."""

    __gtype_name__ = "GraphsPythonStyleEditorBox"

    def __init__(self, window):
        super().__init__(window=window)
        self.params = None

        # Setup Widgets
        for key, _value in STYLE_DICT.items():
            widget = self.get_property(key.replace("-", "_"))
            if isinstance(widget, Adw.EntryRow):
                widget.connect("changed", self._on_entry_change, key)
            elif isinstance(widget, Adw.ComboRow):
                widget.connect("notify::selected", self._on_combo_change, key)
            elif isinstance(widget, Gtk.Scale):
                widget.connect("value-changed", self._on_scale_change, key)
            elif isinstance(widget, Graphs.StyleColorRow):
                widget.connect("notify::color", self._on_color_change, key)
            elif isinstance(widget, Adw.SwitchRow):
                widget.connect("notify::active", self._on_switch_change, key)
            else:
                raise ValueError

        self.props.style_name.connect("changed", self._on_name_change)
        self.props.font_chooser.connect(
            "notify::font-desc",
            self._on_font_change,
        )
        self.props.titlesize.connect(
            "value-changed",
            self._on_titlesize_change,
        )
        self.props.labelsize.connect(
            "value-changed",
            self._on_labelsize_change,
        )
        self.props.color_manager.connect(
            "colors-changed",
            self._on_line_colors_changed,
        )
        self.props.errbar_color_manager.connect(
            "colors-changed",
            self._on_errbar_colors_changed,
        )

    def load_style(self, file: Gio.File) -> None:
        """Load style params from file."""
        self.params = None
        style_params, graphs_params = style_io.parse(
            file,
            Graphs.StyleManager.get_instance().get_system_style_params(),
        )
        stylename = graphs_params["name"]
        self.props.style_name.set_text(stylename)
        for key, value in STYLE_DICT.items():
            value = value[0]
            value = graphs_params[
                value
            ] if value in style_io.STYLE_CUSTOM_PARAMS else style_params[value]
            with contextlib.suppress(KeyError):
                value = VALUE_DICT[key].index(value)
            widget = self.get_property(key.replace("-", "_"))
            if isinstance(widget, Adw.EntryRow):
                widget.set_text(str(value))
            elif isinstance(widget, Adw.ComboRow):
                widget.set_selected(int(value))
            elif isinstance(widget, Gtk.Scale):
                widget.set_value(value)
            elif isinstance(widget, Graphs.StyleColorRow):
                widget.set_color(Graphs.tools_hex_to_rgba(value))
            elif isinstance(widget, Adw.SwitchRow):
                widget.set_active(bool(value))
            else:
                raise ValueError

        # font
        font_description = Pango.FontDescription.new()
        self.font_size = style_params["font.size"]
        font_description.set_size(self.font_size * Pango.SCALE)
        self.props.titlesize.set_value(
            round(style_params["figure.titlesize"] * 2 / self.font_size, 1),
        )
        self.props.labelsize.set_value(
            round(style_params["axes.labelsize"] * 2 / self.font_size, 1),
        )
        font_description.set_family(style_params["font.sans-serif"][0])
        font_description.set_weight(style_params["font.weight"])
        inverted_style_dict = {b: a for a, b in FONT_STYLE_DICT.items()}
        font_description.set_style(
            inverted_style_dict[style_params["font.style"]],
        )
        inverted_variant_dict = {b: a for a, b in FONT_VARIANT_DICT.items()}
        font_description.set_variant(
            inverted_variant_dict[style_params["font.variant"]],
        )
        self.props.font_chooser.set_font_desc(font_description)

        self.check_contrast()

        # line colors
        self.props.color_manager.set_colors(
            style_params["axes.prop_cycle"].by_key()["color"],
        )

        # error bar colors
        self.props.errbar_color_manager.set_colors(
            graphs_params["errorbar.color_cycle"].by_key()["color"],
        )

        self.params, self.graphs_params = style_params, graphs_params

        return stylename

    def save_style(self, file: Gio.File) -> None:
        """Save style params to file."""
        style_io.write(file, self.params, self.graphs_params)

    def _on_line_colors_changed(
        self,
        color_manager: Graphs.StyleColorManager,
    ) -> None:
        """Update line colors in params."""
        if self.params is None:
            return
        line_colors = color_manager.get_colors()
        self.params["axes.prop_cycle"] = cycler(color=line_colors)
        self.params["patch.facecolor"] = line_colors[0]
        self.emit("params-changed")

    def _on_errbar_colors_changed(
        self,
        color_manager: Graphs.StyleColorManager,
    ) -> None:
        """Update errorbar colors in graph-params."""
        if self.graphs_params is None:
            return
        err_colors = color_manager.get_colors()
        self.graphs_params["errorbar.color_cycle"] = cycler(color=err_colors)
        self.emit("params-changed")

    def _on_font_change(self, chooser: Gtk.FontChooser, _param) -> None:
        if self.params is None:
            return
        font_description = chooser.get_font_desc()
        self.params["font.sans-serif"] = [font_description.get_family()]
        self.font_size = font_description.get_size() / Pango.SCALE
        for key in (
            "font.size",
            "xtick.labelsize",
            "ytick.labelsize",
            "legend.fontsize",
            "figure.labelsize",
        ):
            self.params[key] = self.font_size
        font_weight = font_description.get_weight()
        for key in (
            "font.weight",
            "axes.titleweight",
            "axes.labelweight",
            "figure.titleweight",
            "figure.labelweight",
        ):
            self.params[key] = font_weight
        self.params["font.style"] = \
            FONT_STYLE_DICT[font_description.get_style()]
        self.params["font.variant"] = \
            FONT_VARIANT_DICT[font_description.get_variant()]
        self.emit("params-changed")

    def _on_titlesize_change(self, entry: Gtk.Entry) -> None:
        if self.params is None:
            return
        titlesize = round(entry.get_value() / 2 * self.font_size, 1)
        self.params["figure.titlesize"] = titlesize
        self.params["axes.titlesize"] = titlesize
        self.emit("params-changed")

    def _on_labelsize_change(self, entry: Gtk.Entry) -> None:
        if self.params is None:
            return
        labelsize = round(entry.get_value() / 2 * self.font_size, 1)
        self.params["axes.labelsize"] = labelsize
        self.emit("params-changed")

    def _on_name_change(self, entry: Gtk.Entry) -> None:
        if self.params is None:
            return
        self.graphs_params["name"] = entry.get_text()
        self.emit("params-changed")

    def _apply_value(self, key: str, value) -> None:
        if self.params is None:
            return
        with contextlib.suppress(KeyError):
            value = VALUE_DICT[key][value]
        for item in STYLE_DICT[key]:
            if item in style_io.STYLE_CUSTOM_PARAMS:
                self.graphs_params[item] = value
            else:
                self.params[item] = value

        self.emit("params-changed")

    def _on_color_change(
        self,
        row: Graphs.StyleColorRow,
        _param,
        key: str,
    ) -> None:
        self._apply_value(key, Graphs.tools_rgba_to_hex(row.get_color()))
        self.check_contrast()

    def _on_entry_change(self, entry: Gtk.Entry, key: str) -> None:
        self._apply_value(key, str(entry.get_text()))

    def _on_combo_change(
        self,
        comborow: Adw.ComboRow,
        _param,
        key: str,
    ) -> None:
        self._apply_value(key, comborow.get_selected())

    def _on_scale_change(self, scale: Gtk.Scale, key: str) -> None:
        self._apply_value(key, scale.get_value())

    def _on_switch_change(
        self,
        switchrow: Adw.SwitchRow,
        _param,
        key: str,
    ) -> None:
        self._apply_value(key, bool(switchrow.get_active()))



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for data transformations."""
import logging
import re
from gettext import gettext as _

from gi.repository import Gio, Graphs

from graphs import misc, utilities
from graphs.item import DataItem

import numexpr

import numpy

import scipy

import sympy


def perform_operation(window: Graphs.Window, name: str) -> None:
    """Perform an operation."""
    interaction_mode = window.get_mode()
    if name == "cut" and interaction_mode != Graphs.Mode.SELECT:
        return
    args = []
    actions_settings = Graphs.Application.get_settings_child("actions")
    if name in ("center", "smoothen"):
        args = [actions_settings.get_enum(name)]
    if name == "smoothen":
        args.append(actions_settings.get_child(name))
    elif "translate" in name or "multiply" in name:
        args.append(window.get_operation_value(name))

    data = window.get_data()
    figure_settings = data.get_figure_settings()

    if hasattr(CommonOperations, name):
        all_success = getattr(CommonOperations, name)(window)
    else:
        all_success = False
        for item in data:
            if not item.get_selected():
                continue
            if isinstance(item, Graphs.EquationItem):
                operations_class = EquationOperations
            elif isinstance(item, Graphs.DataItem):
                operations_class = DataOperations
            else:
                continue
            success, message = operations_class.execute(
                item,
                name,
                figure_settings,
                interaction_mode,
                *args,
            )
            if message:
                window.add_toast_string(message)
            all_success = success or all_success
    if all_success:
        data.add_history_state()
        data.optimize_limits()


class DataHelper():
    """Helper methods that assist with the handling of the data."""

    @staticmethod
    def get_xydata(
        interaction_mode: Graphs.Mode,
        selected_limits: tuple[float, float],
        item: Graphs.DataItem,
    ) -> tuple[numpy.ndarray, numpy.ndarray]:
        """Get the X and Y data of a DataItem."""
        xdata, ydata = item.get_xydata()
        if interaction_mode != Graphs.Mode.SELECT:
            return xdata, ydata
        startx, stopx = selected_limits
        # If startx and stopx are not out of range, that is,
        # if the item data is within the highlight
        if stopx < min(xdata) or startx > max(xdata):
            return None, None
        mask = numpy.greater_equal(xdata, startx)
        mask &= numpy.less_equal(xdata, stopx)
        return xdata[mask], ydata[mask]

    @staticmethod
    def get_selected_limits(
        figure_settings: Graphs.FigureSettings,
        interaction_mode: Graphs.Mode,
        item: Graphs.DataItem,
    ) -> tuple[float, float]:
        """Get the min and max value of the item within the selected range."""
        if item.get_xposition() == Graphs.XPosition.BOTTOM:
            min_bottom = figure_settings.get_min_bottom()
            max_bottom = figure_settings.get_max_bottom()
            if interaction_mode != Graphs.Mode.SELECT:
                return min_bottom, max_bottom
            scale = figure_settings.get_bottom_scale()
            min_x = Graphs.get_value_at_fraction(
                figure_settings.get_min_selected(),
                min_bottom,
                max_bottom,
                scale,
            )
            max_x = Graphs.get_value_at_fraction(
                figure_settings.get_max_selected(),
                min_bottom,
                max_bottom,
                scale,
            )
        else:
            min_top = figure_settings.get_min_top()
            max_top = figure_settings.get_max_top()
            if interaction_mode != Graphs.Mode.SELECT:
                return min_top, max_top
            scale = figure_settings.get_top_scale()
            min_x = Graphs.get_value_at_fraction(
                figure_settings.get_min_selected(),
                min_top,
                max_top,
                scale,
            )
            max_x = Graphs.get_value_at_fraction(
                figure_settings.get_max_selected(),
                min_top,
                max_top,
                scale,
            )
        return min_x, max_x

    @staticmethod
    def sort_data(
        xdata: numpy.ndarray,
        ydata: numpy.ndarray,
    ) -> tuple[numpy.ndarray, numpy.ndarray]:
        """Sort data."""
        idx = numpy.argsort(xdata)
        return xdata[idx], ydata[idx]


class CommonOperations():
    """Operations to be performed on all kind of items."""

    @staticmethod
    def custom_transformation(window: Graphs.Window) -> bool:
        """Perform a custom operation on the dataset."""

        def on_accept(_dialog, input_x, input_y, discard):
            data = window.get_data()
            figure_settings = data.get_figure_settings()

            for item in data:
                if not item.get_selected():
                    continue
                if isinstance(item, Graphs.EquationItem):
                    operations_class = EquationOperations
                elif isinstance(item, Graphs.DataItem):
                    operations_class = DataOperations
                success, message = operations_class.execute(
                    item,
                    "transform",
                    figure_settings,
                    window.get_mode(),
                    input_x,
                    input_y,
                    discard,
                )
                if message:
                    fail_message = _(
                        "Unable to perform transformation, "
                        "make sure the syntax is correct",
                    )
                    toast = message if success else fail_message
                    window.add_toast_string(toast)

                data.add_history_state()
                data.optimize_limits()

        dialog = Graphs.TransformDialog.new(window)
        dialog.connect("accept", on_accept)
        return False

    @staticmethod
    def combine(window: Graphs.Window) -> bool:
        """Combine the selected data into a new data set."""
        data = window.get_data()
        mode = window.get_mode()
        settings = data.get_figure_settings()

        new_xdata, new_ydata, xerr, yerr = [], [], [], []
        some_x, some_y = False, False

        for item in data:
            if not item.get_selected():
                continue

            lims = DataHelper.get_selected_limits(settings, mode, item)

            if isinstance(item, Graphs.EquationItem):
                eq = item.get_preprocessed_equation()
                xdata, ydata = utilities.equation_to_data(eq, lims)
                xerr, yerr = None, None
            elif isinstance(item, Graphs.DataItem):
                xdata, ydata = DataHelper().get_xydata(mode, lims, item)
                if item.get_xerr() is not None and xerr is not None:
                    xerr.append(item.get_xerr())
                    some_x = True
                else:
                    xerr = None
                if item.get_yerr() is not None and yerr is not None:
                    yerr.append(item.get_yerr())
                    some_y = True
                else:
                    yerr = None

            new_xdata.append(xdata)
            new_ydata.append(ydata)

        if not new_xdata or not new_ydata:
            window.add_toast_string(_("No data found in highlighted area"))
            return False

        if (some_x and xerr is None) or (some_y and yerr is None):
            msg = _("Some items lack error bars; they will be discarded")
            window.add_toast_string(msg)

        new_xdata = numpy.concatenate(new_xdata)
        idx = numpy.argsort(new_xdata)
        data.add_items([
            DataItem.new(
                data.get_selected_style_params(),
                new_xdata[idx],
                numpy.concatenate(new_ydata)[idx],
                xerr=None if xerr is None else numpy.concatenate(xerr)[idx],
                yerr=None if yerr is None else numpy.concatenate(yerr)[idx],
                name=_("Combined Data"),
            ),
        ])
        return True

    @staticmethod
    def shift(window: Graphs.Window) -> None:
        """Shift data."""
        interaction_mode = window.get_mode()
        data = window.get_data()
        figure_settings = data.get_figure_settings()
        data_list = ([
            item for item in data if item.get_selected()
            and isinstance(item, (Graphs.EquationItem, Graphs.DataItem))
        ])
        ranges = [
            figure_settings.get_max_left() - figure_settings.get_min_left(),
            figure_settings.get_max_right() - figure_settings.get_min_right(),
        ]
        left_scale = figure_settings.get_left_scale()
        right_scale = figure_settings.get_right_scale()
        for index, item in enumerate(data_list):
            selected_limits = DataHelper.get_selected_limits(
                figure_settings,
                interaction_mode,
                item,
            )
            startx, stopx = selected_limits
            scale = right_scale if item.get_yposition() else left_scale
            if isinstance(item, Graphs.EquationItem):
                xdata, ydata = utilities.equation_to_data(
                    item.get_preprocessed_equation(),
                    selected_limits,
                )
            elif isinstance(item, Graphs.DataItem):
                xdata, ydata = item.get_xydata()
                if interaction_mode == Graphs.Mode.SELECT:
                    # If startx and stopx are not out of range, that is,
                    # if the item data is within the highlight
                    if stopx < min(xdata) or startx > max(xdata):
                        continue
                    data_mask = numpy.greater_equal(xdata, startx)
                    data_mask &= numpy.less_equal(xdata, stopx)
                    xdata, ydata = xdata[data_mask], ydata[data_mask]
            if min(xdata.size, ydata.size) == 0:
                continue

            shift_value = 0

            item = data_list[0]
            for i in range(index + 1):
                previous_item = item
                item = data_list[i]
                y_range = ranges[item.get_yposition()]

                if isinstance(previous_item, Graphs.EquationItem):
                    prev_min, prev_max = startx, stopx
                else:
                    prev_xdata = previous_item.get_xdata()
                    prev_min, prev_max = min(prev_xdata), max(prev_xdata)
                    if interaction_mode == Graphs.Mode.SELECT:
                        prev_min = max(prev_min, startx)
                        prev_min = min(prev_max, stopx)

                xmin, xmax = min(xdata), max(xdata)
                if xmin >= prev_min and xmax <= prev_max:
                    mask = numpy.greater_equal(xdata, xmin)
                    mask &= numpy.less_equal(xdata, xmax)
                    ydata = ydata[mask]
                nonzero_data = ydata[ydata != 0]
                ymin = nonzero_data.min()
                ymax = nonzero_data.max()

                if scale == Graphs.Scale.LOG:
                    shift_value += \
                        numpy.log10(abs(ymax / ymin)) \
                        + 0.1 * numpy.log10(y_range)
                elif scale == Graphs.Scale.LOG2:
                    shift_value += \
                        numpy.log2(abs(ymax / ymin)) \
                        + 0.1 * numpy.log2(y_range)
                else:
                    shift_value += (ymax - ymin) + 0.1 * y_range
            if shift_value == 0:
                continue
            shift_value = Graphs.math_tools_sig_fig_round(shift_value, 3)
            if isinstance(item, Graphs.EquationItem):
                equation = item.get_preprocessed_equation()
                if scale == Graphs.Scale.LOG:
                    equation = f"({equation})*10**{shift_value}"
                elif scale == Graphs.Scale.LOG2:
                    equation = f"({equation})*2**{shift_value}"
                else:
                    equation = f"{equation}+{shift_value}"
                equation = str(sympy.simplify(equation))
                item.set_equation(Graphs.prettify_equation(equation))
                continue
            if isinstance(item, Graphs.DataItem):
                if scale == Graphs.Scale.LOG:
                    new_ydata = ydata * 10**shift_value
                elif scale == Graphs.Scale.LOG2:
                    new_ydata = ydata * 2**shift_value
                else:  # Apply linear scaling
                    new_ydata = ydata + shift_value
                if interaction_mode == Graphs.Mode.SELECT:
                    item_ydata = item.get_ydata().copy()
                    item_ydata[data_mask] = new_ydata
                    new_ydata = item_ydata
                item.set_xydata((item.get_xdata(), new_ydata))
                continue
        return True


XDATA = numpy.linspace(0, 10, 10)


class EquationOperations():
    """Operations to be performed on equation items."""

    @staticmethod
    def execute(
        item: Graphs.EquationItem,
        name: str,
        figure_settings: Graphs.FigureSettings,
        _interaction_mode: Graphs.Mode,
        *args,
    ) -> tuple[bool, str]:
        """Execute the operation on the given item."""
        old_limits = figure_settings.get_limits().values()
        try:
            callback = getattr(EquationOperations, name)
            if name in ("normalize", "center", "transform"):
                args = [(
                    old_limits[item.get_xposition()],
                    old_limits[item.get_yposition() + 1],
                )] + list(args)
            equation = item.get_preprocessed_equation()
            equation = str(sympy.simplify(callback(equation, *args)))
            try:
                numexpr.evaluate(equation, local_dict={"x": XDATA})
            except (KeyError, SyntaxError, ValueError, TypeError) as e:
                raise misc.InvalidEquationError(
                    _(
                        "The operation on {name} "
                        "did not result in a plottable equation",
                    ).format(name=item.get_name()),
                ) from e
            item.set_equation(Graphs.prettify_equation(equation))
        except misc.InvalidEquationError as error:
            return False, error.message
        except (NotImplementedError, AttributeError, KeyError):
            return False, _("Operation not supported for equations.")
        return True, ""

    @staticmethod
    def translate_x(equation, offset) -> str:
        """Translate all selected data on the x-axis."""
        return re.sub(r"(?<!e)x(?!p)", f"(x+{offset})", equation)

    @staticmethod
    def translate_y(equation, offset) -> str:
        """Translate all selected data on the y-axis."""
        return f"({equation})+{offset}"

    @staticmethod
    def multiply_x(equation, multiplier: float) -> str:
        """Multiply all selected data on the x-axis."""
        return re.sub(r"(?<!e)x(?!p)", f"(x*{multiplier})", equation)

    @staticmethod
    def multiply_y(equation, multiplier: float) -> str:
        """Multiply all selected data on the y-axis."""
        return f"({equation})*{multiplier}"

    @staticmethod
    def normalize(equation, limits) -> str:
        """Normalize all selected data."""
        ydata = utilities.equation_to_data(equation, limits)[1]
        return f"({equation})/{max(ydata)}"

    @staticmethod
    def center(equation, limits, center_maximum: int) -> str:
        """
        Center all selected data.

        Depending on the key, will center either on the middle coordinate, or
        on the maximum value of the data
        """
        xdata, ydata = utilities.equation_to_data(equation, limits)
        if center_maximum == 0:  # Center at maximum Y
            x = misc.X
            equation = sympy.sympify(equation)
            derivative = sympy.diff(equation, x)
            critical_points = sympy.solveset(
                derivative,
                x,
                domain=sympy.Interval(limits[0], limits[1]),
            )
            endpoints = \
                [equation.subs(x, limits[0]), equation.subs(x, limits[1])]

            try:
                critical_values = \
                    [equation.subs(x, cp) for cp in critical_points]
                if critical_values:
                    max_index = critical_values.index(max(critical_values))
                    middle_value = list(critical_points)[max_index]
                else:
                    max_index = endpoints.index(max(endpoints))
                    middle_value = [limits[0], limits[1]][max_index]

            # If we don't manage to solve this analytically, just find
            # the maximum by calculating
            except TypeError:
                middle_index = numpy.argmax(ydata)
                middle_value = xdata[middle_index]

        elif center_maximum == 1:  # Center at middle
            middle_value = (min(xdata) + max(xdata)) / 2
        return re.sub(r"(?<!e)x(?!p)", f"(x+{middle_value})", equation)

    @staticmethod
    def derivative(equation) -> str:
        """Calculate derivative of all selected data."""
        return str(sympy.diff(equation, misc.X))

    @staticmethod
    def integral(equation) -> str:
        """Calculate indefinite integral of all selected data."""
        return str(sympy.integrate(equation, misc.X))

    @staticmethod
    def fft(equation) -> str:
        """Perform Fourier transformation on all selected data."""
        k = sympy.Symbol("k")
        equation = str(sympy.fourier_transform(equation, misc.X, k))
        return equation.replace("k", "x")

    @staticmethod
    def inverse_fft(equation) -> str:
        """Perform Inverse Fourier transformation on all selected data."""
        k = sympy.Symbol("k")
        equation = str(sympy.fourier_transform(equation, misc.X, k))
        return equation.replace("k", "x")

    @staticmethod
    def transform(
        equation: str,
        limits: list,
        input_x: str,
        input_y: str,
        _discard: bool,
    ) -> str:
        """Perform custom transformation."""
        xdata, ydata = utilities.equation_to_data(equation, limits)
        local_dict = {
            "x": xdata,
            "y": ydata,
            "x_min": min(xdata),
            "x_max": max(xdata),
            "y_min": min(ydata),
            "y_max": max(ydata),
            "counts": len(xdata),
            "x_mean": numpy.mean(xdata),
            "y_mean": numpy.mean(ydata),
            "x_std": numpy.std(xdata),
            "y_std": numpy.std(ydata),
            "x_median": numpy.median(xdata),
            "y_median": numpy.median(ydata),
            "x_sum": sum(xdata),
            "y_sum": sum(ydata),
        }

        for key, value in local_dict.items():
            if key not in ("x", "y"):
                input_x = input_x.lower().replace(key, str(value))
                input_y = input_y.lower().replace(key, str(value))

        equation = re.sub(r"(?<!e)x(?!p)", input_x, equation)
        return input_y.lower().replace("y", equation)


_return = (numpy.ndarray, numpy.ndarray, bool, bool)


class DataOperations():
    """Operations to be performed on data items."""

    @staticmethod
    def execute(
        item: Graphs.DataItem,
        name: str,
        figure_settings: Graphs.FigureSettings,
        interaction_mode: Graphs.Mode,
        *args,
    ) -> tuple[bool, str]:
        """Execute the operation on the given item."""
        xdata, ydata = item.get_xydata()
        if interaction_mode == Graphs.Mode.SELECT:
            startx, stopx = DataHelper.get_selected_limits(
                figure_settings,
                interaction_mode,
                item,
            )
            # If startx and stopx are not out of range, that is,
            # if the item data is within the highlight
            xmin = min(xdata)
            if not (startx < xmin and stopx < xmin or (startx > max(xdata))):
                mask = numpy.greater_equal(xdata, startx)
                mask &= numpy.less_equal(xdata, stopx)
                xdata, ydata = xdata[mask], ydata[mask]
            else:
                xdata, ydata = None, None
        if not (xdata is not None and len(xdata) != 0):
            return False, _("No data found within the highlighted area")
        try:
            callback = getattr(DataOperations, name)
            message = ""
            new_xdata, new_ydata, sort, discard = callback(
                item, xdata, ydata, *args,
            )
        except NotImplementedError:
            return False, _("Operation not supported for data items")
        # May run into this exception for custom transformations:
        except (RuntimeError, ValueError, KeyError, SyntaxError) as exception:
            message = _("{name}: Error performing the operation")
            return False, message.format(name=exception.__class__.__name__)
        xerr = item.get_xerr()
        yerr = item.get_yerr()
        if interaction_mode == Graphs.Mode.SELECT:
            if discard:
                logging.debug("Discard is true")
                message = _(
                    "Data that was outside of the highlighted area has"
                    " been discarded",
                )
            elif new_xdata is None:  # If cut action was performed
                new_xdata = item.get_xdata()[~mask]
                new_ydata = item.get_ydata()[~mask]
                xerr = xerr[~mask] if xerr is not None else None
                yerr = yerr[~mask] if yerr is not None else None
            else:
                logging.debug("Discard is false")
                xdata = item.get_xdata().copy()
                ydata = item.get_ydata().copy()
                xdata[mask] = new_xdata
                ydata[mask] = new_ydata
                new_xdata, new_ydata = xdata, ydata
        if sort:
            logging.debug("Sorting data")
            new_xdata, new_ydata = DataHelper.sort_data(new_xdata, new_ydata)
        item.set_data_tuple((new_xdata, new_ydata, xerr, yerr))
        return True, message

    @staticmethod
    def translate_x(_item, xdata: list, ydata: list, offset: float) -> _return:
        """
        Translate all selected data on the x-axis.

        Amount to be shifted is equal to the value in the translate_x entry
        widget.
        Will show a toast if a ValueError is raised, typically when a user
        entered an invalid number (e.g. comma instead of point separators)
        """
        return xdata + offset, ydata, True, False

    @staticmethod
    def translate_y(_item, xdata: list, ydata: list, offset: float) -> _return:
        """
        Translate all selected data on the y-axis.

        Amount to be shifted is equal to the value in the translate_y entry
        widget.
        Will show a toast if a ValueError is raised, typically when a user
        entered an invalid number (e.g. comma instead of point separators)
        """
        return xdata, ydata + offset, False, False

    @staticmethod
    def multiply_x(
        _item,
        xdata: list,
        ydata: list,
        multiplier: float,
    ) -> _return:
        """
        Multiply all selected data on the x-axis.

        Amount to be multiplied is equal to the value in the multiply_x entry
        widget
        Will show a toast if a ValueError is raised, typically when a user
        entered an invalid number (e.g. comma instead of point separators)
        """
        return xdata * multiplier, ydata, True, False

    @staticmethod
    def multiply_y(
        _item,
        xdata: list,
        ydata: list,
        multiplier: float,
    ) -> _return:
        """
        Multiply all selected data on the y-axis.

        Amount to be multiplied is equal to the value in the multiply_y entry
        widget
        Will show a toast if a ValueError is raised, typically when a user
        entered an invalid number (e.g. comma instead of point separators)
        """
        return xdata, ydata * multiplier, False, False

    @staticmethod
    def normalize(_item, xdata: list, ydata: list) -> _return:
        """Normalize all selected data."""
        return xdata, ydata / max(ydata), False, False

    @staticmethod
    def smoothen(
        _item,
        xdata: list,
        ydata: list,
        smooth_type: int,
        settings: Gio.Settings,
    ) -> _return:
        """Smoothen y-data."""
        if smooth_type == 0:
            minimum = settings.get_int("savgol-polynomial") + 1
            window_percentage = settings.get_int("savgol-window") / 100
            window = max(minimum, int(len(xdata) * window_percentage))
            new_ydata = scipy.signal.savgol_filter(
                ydata,
                window,
                settings.get_int("savgol-polynomial"),
            )
        elif smooth_type == 1:
            box_points = settings.get_int("moving-average-box")
            box = numpy.ones(box_points) / box_points
            new_ydata = numpy.convolve(ydata, box, mode="same")
        return xdata, new_ydata, False, False

    @staticmethod
    def center(
        _item,
        xdata: list,
        ydata: list,
        center_maximum: int,
    ) -> _return:
        """
        Center all selected data.

        Depending on the key, will center either on the middle coordinate, or
        on the maximum value of the data
        """
        if center_maximum == 0:  # Center at maximum Y
            middle_index = numpy.argmax(ydata)
            middle_value = xdata[middle_index]
        elif center_maximum == 1:  # Center at middle
            middle_value = (min(xdata) + max(xdata)) / 2
        return xdata - middle_value, ydata, True, False

    @staticmethod
    def cut(_item, _xdata, _ydata) -> _return:
        """Cut selected data over the span that is selected."""
        return None, None, False, False

    @staticmethod
    def derivative(_item, xdata: list, ydata: list) -> _return:
        """Calculate derivative of all selected data."""
        return xdata, numpy.gradient(ydata, xdata), False, True

    @staticmethod
    def integral(_item, xdata: list, ydata: list) -> _return:
        """Calculate indefinite integral of all selected data."""
        indefinite_integral = scipy.integrate.cumulative_trapezoid(
            ydata,
            xdata,
            initial=0,
        )
        return xdata, indefinite_integral, False, True

    @staticmethod
    def fft(_item, xdata: list, ydata: list) -> _return:
        """Perform Fourier transformation on all selected data."""
        y_fourier = numpy.fft.fft(ydata)
        x_fourier = numpy.fft.fftfreq(len(xdata), xdata[1] - xdata[0])
        y_fourier = [value.real for value in y_fourier]
        return x_fourier, y_fourier, False, True

    @staticmethod
    def inverse_fft(_item, xdata: list, ydata: list) -> _return:
        """Perform Inverse Fourier transformation on all selected data."""
        y_fourier = numpy.fft.ifft(ydata)
        x_fourier = numpy.fft.fftfreq(len(xdata), xdata[1] - xdata[0])
        y_fourier = [value.real for value in y_fourier]
        return x_fourier, y_fourier, False, True

    @staticmethod
    def transform(
        _item,
        xdata: list,
        ydata: list,
        input_x: str,
        input_y: str,
        discard: bool = False,
    ) -> _return:
        """Perform custom transformation."""
        local_dict = {
            "x": xdata,
            "y": ydata,
            "x_min": min(xdata),
            "x_max": max(xdata),
            "y_min": min(ydata),
            "y_max": max(ydata),
            "counts": len(xdata),
            "x_mean": numpy.mean(xdata),
            "y_mean": numpy.mean(ydata),
            "x_std": numpy.std(xdata),
            "y_std": numpy.std(ydata),
            "x_median": numpy.median(xdata),
            "y_median": numpy.median(ydata),
            "x_sum": sum(xdata),
            "y_sum": sum(ydata),
        }
        # Add array of zeros to return values, such that output remains a list
        # of the correct size, even when a float is given as input.
        return (
            numexpr.evaluate(
                Graphs.preprocess_equation(input_x) + "+ 0*x",
                local_dict,
            ),
            numexpr.evaluate(
                Graphs.preprocess_equation(input_y) + "+ 0*y",
                local_dict,
            ),
            True,
            discard,
        )



# SPDX-License-Identifier: GPL-3.0-or-later
"""Python Helper - Python part."""
from gi.repository import Gio, Graphs

from graphs import curve_fitting, export_items, file_io, operations, utilities
from graphs.figure import Figure
from graphs.style_editor import PythonStyleEditor
from graphs.window import PythonWindow

import numexpr

import numpy

import sympy

_REQUESTS = (
    "create-style-editor",
    "create-window",
    "curve-fitting-dialog",
    "equation-to-data",
    "evaluate-expression",
    "export-figure",
    "export-items",
    "perform-operation",
    "python-method",
    "simplify-equation",
    "validate-equation",
)

XDATA = numpy.linspace(0, 10, 10)


class PythonHelper(Graphs.PythonHelper):
    """Python helper for python only calls."""

    def __init__(self):
        super().__init__()
        for request in _REQUESTS:
            request = request + "-request"
            self.connect(
                request,
                getattr(self, "_on_" + request.replace("-", "_")),
            )

    @staticmethod
    def _on_create_style_editor_request(self) -> Graphs.StyleEditor:
        return PythonStyleEditor()

    @staticmethod
    def _on_create_window_request(self) -> Graphs.Window:
        return PythonWindow()

    @staticmethod
    def _on_curve_fitting_dialog_request(
        self,
        window: Graphs.Window,
        item: Graphs.Item,
    ) -> None:
        return curve_fitting.CurveFittingDialog(window, item)

    @staticmethod
    def _on_equation_to_data_request(
        self,
        equation: str,
        xstart: float,
        xstop: float,
        steps: int,
        scale: Graphs.Scale,
    ) -> Graphs.DataHolder:
        xdata, ydata = utilities.equation_to_data(
            equation,
            [xstart, xstop],
            steps,
            scale,
        )
        return Graphs.DataHolder.new(xdata, ydata, None, None)

    @staticmethod
    def _on_evaluate_expression_request(
        self,
        equation: str,
        steps: int,
        var: str,
    ) -> bool:
        local_dict = {var: numpy.arange(steps)}
        try:
            data = numexpr.evaluate(equation, local_dict=local_dict)
            if data.ndim == 0:
                data = numpy.full(steps, data)
            self.set_evaluate_expression_result(data)
            return True
        except (KeyError, SyntaxError, ValueError, TypeError):
            return False

    @staticmethod
    def _on_export_items_request(
        self,
        window: Graphs.Window,
        mode: str,
        file: Gio.File,
        items: list[Graphs.Item],
        _n_items: int,
    ) -> None:
        figure_settings = window.get_data().get_figure_settings()
        return export_items.export_items(mode, file, items, figure_settings)

    @staticmethod
    def _on_export_figure_request(
        self,
        file: Gio.File,
        settings: Gio.Settings,
        data: Graphs.Data,
    ) -> None:
        with file_io.open(file, "wb") as file_like:
            figure = Figure(
                data.get_selected_style_params(),
                data,
                figure_settings=data.get_figure_settings(),
            )

            vector_formats = ["pdf", "eps", "ps", "svg"]
            fmt = settings.get_string("file-format")
            dpi = 100 if fmt.lower() in vector_formats else figure.get_dpi()
            width_inches = settings.get_int("width") / dpi
            height_inches = settings.get_int("height") / dpi
            figure.set_size_inches(width_inches, height_inches)
            figure.savefig(
                file_like,
                format=fmt,
                dpi=dpi,
                transparent=settings.get_boolean("transparent"),
                bbox_inches=None,
            )

    @staticmethod
    def _on_perform_operation_request(
        self,
        window: Graphs.Window,
        name: str,
    ) -> None:
        operations.perform_operation(window, name)

    @staticmethod
    def _on_python_method_request(self, obj, method: str) -> None:
        getattr(obj, method)()

    @staticmethod
    def _on_simplify_equation_request(self, equation: str) -> str:
        return str(sympy.simplify(equation))

    @staticmethod
    def _on_validate_equation_request(self, equation: str) -> bool:
        try:
            ydata = numexpr.evaluate(equation, local_dict={"x": XDATA})
            return numpy.isfinite(ydata).any()
        except (KeyError, SyntaxError, ValueError, TypeError):
            return False



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for file operations."""
import json
from xml.dom import minidom

from gi.repository import Gio

import gio_pyio


def open(*args, **kwargs):
    """
    Python open wrapper.

    Currently based on gio_pyio, but may change in the future.
    """
    return gio_pyio.open(*args, **kwargs)


def parse_json(file: Gio.File) -> dict:
    """Parse a json file to a python dict."""
    with gio_pyio.open(file, "rb") as wrapper:
        return json.load(wrapper)


def write_json(file: Gio.File, json_object: dict, pretty_print=True) -> None:
    """Write a python dict to a python file."""
    with gio_pyio.open(file, "wt") as wrapper:
        json.dump(
            json_object,
            wrapper,
            indent=4 if pretty_print else None,
            sort_keys=True,
        )


def parse_xml(file: Gio.File) -> dict:
    """Parse a xml file to a python dict."""
    with gio_pyio.open(file, "rb") as wrapper:
        return minidom.parse(wrapper)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for parsing sql files."""
from gettext import pgettext as C_

from gi.repository import GLib, Graphs, Gtk

from graphs.file_import.parsers import Parser
from graphs.misc import ParseError


class SqlParser(Parser):
    """SQL database parser."""

    __gtype_name__ = "GraphsSqlParser"

    def __init__(self):
        super().__init__(
            "sql",
            C_("import-mode", "Database"),
            C_("file-filter", "SQLite Database"),
            ["db", "sqlite", "sqlite3"],
        )

    @staticmethod
    def parse(
        items: Graphs.ItemList,
        settings: Graphs.ImportSettings,
        data: Graphs.Data,
    ) -> None:
        """Import data from sqlite database file."""
        db_reader = settings.get_item("db-reader")
        msg = db_reader.parse(items, settings, data)
        if msg != "":
            raise ParseError(msg)

    @staticmethod
    def init_settings(settings: Graphs.ImportSettings) -> bool:
        """Init settings with default table and column selection."""
        try:
            db_reader = Graphs.DatabaseReader.new(settings)
            db_reader.set_default_selection()
            settings.set_item("db-reader", db_reader)
            return True
        except GLib.GError:
            return False

    @staticmethod
    def init_settings_widgets(
        settings: Graphs.ImportSettings,
        box: Gtk.Box,
    ) -> None:
        """Append SQL-specific settings widgets."""
        if not settings.get_item("db-reader"):
            return
        box.append(Graphs.SqlGroup.new(settings))



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for "parsing" project files."""
from gettext import pgettext as C_

from gi.repository import Graphs

from graphs import project
from graphs.file_import.parsers import Parser
from graphs.item import ItemFactory
from graphs.misc import ParseError
from graphs.project import ProjectParseError


class ProjectParser(Parser):
    """Project parser."""

    __gtype_name__ = "GraphsProjectParser"

    def __init__(self):
        super().__init__(
            "project",
            C_("import-mode", "Project"),
            C_("file-filter", "Graphs Project File"),
            ["graphs"],
        )

    @staticmethod
    def parse(
        items: Graphs.ItemList,
        settings: Graphs.ImportSettings,
        _data: Graphs.Data,
    ) -> None:
        """Import data from project file."""
        try:
            project_dict = project.read_project_file(settings.get_file())
            items = list(map(ItemFactory.new_from_dict, project_dict["data"]))
            items.add_all(items)
        except ProjectParseError as e:
            raise ParseError(e.message) from e



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Custom figure implementation.

Acts as an interface between matplotlib and GObject.
"""
import contextlib
import logging

from gi.repository import GObject, Gio, Graphs, Gtk

from graphs import artist, misc

from matplotlib import RcParams, figure, pyplot


def _ellipsize(text: str, max_chars: int) -> str:
    """Truncate text with an ellipsis."""
    return (text[:max_chars] + "…") if len(text) > max_chars else text


class Figure(GObject.Object, figure.Figure):
    """Custom Figure."""

    __gtype_name__ = "GraphsFigure"

    def __init__(
        self,
        style_params: tuple[RcParams, dict],
        items: Gio.ListModel,
        parent=None,
        figure_settings: Graphs.FigureSettings = None,
    ):
        GObject.Object.__init__(self)
        self._style_params = style_params
        self._items = items
        self.parent = parent
        pyplot.rcParams.update(self._style_params[0])  # apply style_params
        figure.Figure.__init__(self, tight_layout=True)

        self.axis = self.add_subplot(111)
        self.top_left_axis = self.axis.twiny()
        self.right_axis = self.axis.twinx()
        self.top_right_axis = self.top_left_axis.twinx()
        self._legend_axis = self.axis
        self._artists = []

        if figure_settings is not None:
            bottom_scale = \
                Graphs.scale_to_string(figure_settings.get_bottom_scale())
            top_scale = \
                Graphs.scale_to_string(figure_settings.get_top_scale())
            left_scale = \
                Graphs.scale_to_string(figure_settings.get_left_scale())
            right_scale = \
                Graphs.scale_to_string(figure_settings.get_right_scale())
            self.axis.set_xscale(bottom_scale)
            self.axis.set_yscale(left_scale)
            self.top_left_axis.set_xscale(top_scale)
            self.top_left_axis.set_yscale(left_scale)
            self.right_axis.set_xscale(bottom_scale)
            self.right_axis.set_yscale(right_scale)
            self.top_right_axis.set_xscale(top_scale)
            self.top_right_axis.set_yscale(right_scale)

            limits = figure_settings.get_limits().values()
            self.axis.set_xlim(limits[0], limits[1])
            self.axis.set_ylim(limits[4], limits[5])
            self.top_left_axis.set_xlim(limits[2], limits[3])
            self.top_left_axis.set_ylim(limits[4], limits[5])
            self.right_axis.set_xlim(limits[0], limits[1])
            self.right_axis.set_ylim(limits[6], limits[7])
            self.top_right_axis.set_xlim(limits[2], limits[3])
            self.top_right_axis.set_ylim(limits[6], limits[7])

            title = figure_settings.get_title()
            self.axis.set_title(title, picker=True).id = "title"
            self.axis.set_xlabel(
                figure_settings.get_bottom_label(),
                picker=True,
            ).id = "bottom_label"
            self.top_left_axis.set_xlabel(
                figure_settings.get_top_label(),
                picker=True,
            ).id = "top_label"
            self.axis.set_ylabel(
                figure_settings.get_left_label(),
                picker=True,
            ).id = "left_label"
            self.right_axis.set_ylabel(
                figure_settings.get_right_label(),
                picker=True,
            ).id = "right_label"

            self._hide_unselected = figure_settings.get_hide_unselected()
            self._legend = figure_settings.get_legend()
            pos = figure_settings.get_legend_position()
            self._legend_position = Graphs.legend_position_to_string(pos)
        else:
            self._hide_unselected = False
            self._legend = True
            self._legend_position = "best"

        items.connect("items-changed", self._redraw)
        if isinstance(items, Gtk.SelectionModel):
            items.connect("selection-changed", self._redraw)
        self._redraw()

    def _redraw(self, *_args) -> None:
        logging.debug("redrawing figure")
        # bottom, top, left, right
        used_axes = [False, False, False, False]
        visible_axes = [False, False, False, False]
        drawable_items = [x for x in self._items if x.get_selected()] \
            if self._hide_unselected else list(self._items)
        for item in drawable_items:
            xposition = item.get_xposition()
            yposition = item.get_yposition()
            visible_axes[xposition] = True
            visible_axes[2 + yposition] = True
            used_axes[xposition + 2 * yposition] = True
        axes_directions = (
            ("bottom", "left"),  # axis
            ("top", "left"),  # top_left_axis
            ("bottom", "right"),  # right_axis
            ("top", "right"),  # top_right_axis
        )

        if not any(visible_axes):
            visible_axes = (True, False, True, False)  # Left and bottom
            used_axes = (True, False, False, False)  # self.axis visible
            self._legend_axis = self.axis

        params, graphs_params = self._style_params
        draw_frame = params["axes.spines.bottom"]
        ticks = "both" if params["xtick.minor.visible"] else "major"
        for directions, axis, used \
                in zip(axes_directions, self.axes, used_axes):
            axis.get_xaxis().set_visible(False)
            axis.get_yaxis().set_visible(False)
            # Set tick where requested, as long as that axis is not occupied
            # and visible
            if (
                params[f"xtick.{directions[0]}"]
                or params[f"ytick.{directions[1]}"]
            ):
                tick_params = {}
                for i, direction in enumerate(misc.DIRECTIONS):
                    tick_shown = (
                        (draw_frame and not visible_axes[i])
                        or direction in directions
                    ) and params[f"{'x' if i < 2 else 'y'}tick.{direction}"]

                    tick_params[direction] = tick_shown
                    if graphs_params["ticklabels"]:
                        tick_params[f"label{direction}"] = tick_shown

                axis.tick_params(which=ticks, **tick_params)

            for handle in [*axis.collections, *axis.lines, *axis.texts]:
                handle.remove()
            axis_legend = axis.get_legend()
            if axis_legend is not None:
                axis_legend.remove()
            for direction in misc.DIRECTIONS:
                axis.spines[direction].set_visible(
                    direction in directions and used or draw_frame,
                )
            if used:
                self._legend_axis = axis

        self.axis.get_xaxis().set_visible(visible_axes[0])
        self.top_left_axis.get_xaxis().set_visible(visible_axes[1])
        self.axis.get_yaxis().set_visible(visible_axes[2])
        self.right_axis.get_yaxis().set_visible(visible_axes[3])

        self._artists = [
            artist.new_for_item(self, item)
            for item in reversed(drawable_items)
        ]
        self.update_legend()

    def update_legend(self, *_args) -> None:
        """Update the legend or hide if not used."""
        if self._legend and self._artists:
            handles = [
                handle.get_artist() for handle in self._artists
                if handle.legend
            ]
            if handles:
                max_char = max(10, int(self.bbox.width / 15))
                labels = [_ellipsize(h.get_label(), max_char) for h in handles]
                self._legend_axis.legend(
                    handles=handles,
                    labels=labels,
                    loc=self._legend_position,
                    frameon=True,
                    reverse=True,
                )
                self.queue_draw()
                return
        legend = self._legend_axis.get_legend()
        if legend is not None:
            legend.remove()
        self.queue_draw()

    def set_properties(*args):
        """Pass method on to Figure."""
        figure.Figure.set_properties(*args)

    def set_property(*args):
        """Pass method on to Figure."""
        figure.Figure.set_property(*args)

    def queue_draw(self) -> None:
        """Queue a draw when in a canvas."""
        with contextlib.suppress(AttributeError):
            self.canvas.queue_draw()

    @GObject.Property(type=bool, default=False)
    def hide_unselected(self) -> bool:
        """Whether or not to hide unselected items."""
        return self._hide_unselected

    @hide_unselected.setter
    def hide_unselected(self, hide_unselected: bool) -> None:
        self._hide_unselected = hide_unselected
        self._redraw()

    @GObject.Property(type=bool, default=True)
    def legend(self) -> bool:
        """Whether or not, the legend is visible."""
        return self._legend

    @legend.setter
    def legend(self, legend: bool) -> None:
        self._legend = legend
        self.update_legend()

    @GObject.Property(
        type=Graphs.LegendPosition,
        default=Graphs.LegendPosition.BEST,
    )
    def legend_position(self) -> int:
        """Legend Position."""
        return Graphs.LegendPosition.from_string(self._legend_position)

    @legend_position.setter
    def legend_position(self, pos: Graphs.LegendPosition) -> None:
        self._legend_position = Graphs.legend_position_to_string(pos)
        self.update_legend()

    @GObject.Property(type=str)
    def title(self) -> str:
        """Figure title."""
        return self.axis.get_title()

    @title.setter
    def title(self, title: str) -> None:
        self.axis.set_title(title, picker=True).id = "title"
        self.queue_draw()

    @GObject.Property(type=str)
    def bottom_label(self) -> str:
        """Label of the bottom axis."""
        return self.axis.get_xlabel()

    @bottom_label.setter
    def bottom_label(self, label: str) -> None:
        self.axis.set_xlabel(label, picker=True).id = "bottom_label"
        self.queue_draw()

    @GObject.Property(type=str)
    def left_label(self) -> str:
        """Label of the left axis."""
        return self.axis.get_ylabel()

    @left_label.setter
    def left_label(self, label: str) -> None:
        self.axis.set_ylabel(label, picker=True).id = "left_label"
        self.queue_draw()

    @GObject.Property(type=str)
    def top_label(self) -> str:
        """Label of the top axis."""
        return self.top_left_axis.get_xlabel()

    @top_label.setter
    def top_label(self, label: str) -> None:
        self.top_left_axis.set_xlabel(label, picker=True).id = "top_label"
        self.queue_draw()

    @GObject.Property(type=str)
    def right_label(self) -> str:
        """Label of the right axis."""
        return self.right_axis.get_ylabel()

    @right_label.setter
    def right_label(self, label: str) -> None:
        self.right_axis.set_ylabel(label, picker=True).id = "right_label"
        self.queue_draw()

    @GObject.Property(type=int)
    def bottom_scale(self) -> int:
        """Scale of the bottom axis."""
        return Graphs.scale_from_string(self.axis.get_xscale())

    @bottom_scale.setter
    def bottom_scale(self, scale: int) -> None:
        scale = Graphs.scale_to_string(scale)
        for axis in (self.axis, self.right_axis):
            axis.set_xscale(scale)
            axis.set_xlim(None, None)
        self.queue_draw()

    @GObject.Property(type=int)
    def left_scale(self) -> int:
        """Scale of the left axis."""
        return Graphs.scale_from_string(self.axis.get_yscale())

    @left_scale.setter
    def left_scale(self, scale: int) -> None:
        scale = Graphs.scale_to_string(scale)
        for axis in (self.axis, self.top_left_axis):
            axis.set_yscale(scale)
            axis.set_ylim(None, None)
        self.queue_draw()

    @GObject.Property(type=int)
    def top_scale(self) -> int:
        """Scale of the top axis."""
        return Graphs.scale_from_string(self.top_left_axis.get_xscale())

    @top_scale.setter
    def top_scale(self, scale: int) -> None:
        scale = Graphs.scale_to_string(scale)
        for axis in (self.top_right_axis, self.top_left_axis):
            axis.set_xscale(scale)
            axis.set_xlim(None, None)
        self.queue_draw()

    @GObject.Property(type=int)
    def right_scale(self) -> int:
        """Scale of the right axis."""
        return Graphs.scale_from_string(self.right_axis.get_yscale())

    @right_scale.setter
    def right_scale(self, scale: int) -> None:
        scale = Graphs.scale_to_string(scale)
        for axis in (self.top_right_axis, self.right_axis):
            axis.set_yscale(scale)
            axis.set_ylim(None, None)
        self.queue_draw()

    @GObject.Property(type=float)
    def min_bottom(self) -> float:
        """Lower limit for the bottom axis."""
        return self.axis.get_xlim()[0]

    @min_bottom.setter
    def min_bottom(self, value: float) -> None:
        for axis in (self.axis, self.right_axis):
            axis.set_xlim(value, None)
        self.queue_draw()

    @GObject.Property(type=float)
    def max_bottom(self) -> float:
        """Upper limit for the bottom axis."""
        return self.axis.get_xlim()[1]

    @max_bottom.setter
    def max_bottom(self, value: float) -> None:
        for axis in (self.axis, self.right_axis):
            axis.set_xlim(None, value)
        self.queue_draw()

    @GObject.Property(type=float)
    def min_left(self) -> float:
        """Lower limit for the left axis."""
        return self.axis.get_ylim()[0]

    @min_left.setter
    def min_left(self, value: float) -> None:
        for axis in (self.axis, self.top_left_axis):
            axis.set_ylim(value, None)
        self.queue_draw()

    @GObject.Property(type=float)
    def max_left(self) -> float:
        """Upper limit for the left axis."""
        return self.axis.get_ylim()[1]

    @max_left.setter
    def max_left(self, value: float) -> None:
        for axis in (self.axis, self.top_left_axis):
            axis.set_ylim(None, value)
        self.queue_draw()

    @GObject.Property(type=float)
    def min_top(self) -> float:
        """Lower limit for the top axis."""
        return self.top_left_axis.get_xlim()[0]

    @min_top.setter
    def min_top(self, value: float) -> None:
        for axis in (self.top_left_axis, self.top_right_axis):
            axis.set_xlim(value, None)
        self.queue_draw()

    @GObject.Property(type=float)
    def max_top(self) -> float:
        """Upper limit for the top axis."""
        return self.top_left_axis.get_xlim()[1]

    @max_top.setter
    def max_top(self, value: float) -> None:
        for axis in (self.top_left_axis, self.top_right_axis):
            axis.set_xlim(None, value)
        self.queue_draw()

    @GObject.Property(type=float)
    def min_right(self) -> float:
        """Lower limit for the right axis."""
        return self.right_axis.get_ylim()[0]

    @min_right.setter
    def min_right(self, value: float) -> None:
        for axis in (self.right_axis, self.top_right_axis):
            axis.set_ylim(value, None)
        self.queue_draw()

    @GObject.Property(type=float)
    def max_right(self) -> float:
        """Upper limit for the right axis."""
        return self.right_axis.get_ylim()[1]

    @max_right.setter
    def max_right(self, value: float) -> None:
        for axis in (self.right_axis, self.top_right_axis):
            axis.set_ylim(None, value)
        self.queue_draw()



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for migrating old data to new structures."""
import contextlib
import logging
import pickle
import sys

from gi.repository import Gio

import gio_pyio

LIMITS = [
    "min_bottom",
    "max_bottom",
    "min_top",
    "max_top",
    "min_left",
    "max_left",
    "min_right",
    "max_right",
]

ITEM_MIGRATION_TABLE = {
    "plot_x_position": "xposition",
    "plot_y_position": "yposition",
    "x_anchor": "xanchor",
    "y_anchor": "yanchor",
    "key": "uuid",
}

ITEM_VALUE_MIGRATION_TABLE = {
    "linestyle": ["none", "solid", "dotted", "dashed", "dashdot"],
    "markerstyle": [
        "none",
        ".",
        ",",
        "o",
        "v",
        "^",
        "<",
        ">",
        "8",
        "s",
        "p",
        "*",
        "h",
        "H",
        "+",
        "x",
        "D",
        "d",
        "|",
        "_",
        "P",
        "X",
    ],
    "xposition": ["bottom", "top"],
    "yposition": ["left", "right"],
}

PLOT_SETTINGS_MIGRATION_TABLE = {
    "xlabel": "bottom_label",
    "ylabel": "left_label",
    "xscale": "bottom_scale",
    "yscale": "left_scale",
    "use_custom_plot_style": "use_custom_style",
    "custom_plot_style": "custom_style",
    "mix_right": "min_right",
}

LEGEND_POSITIONS = [
    "best",
    "upper right",
    "upper left",
    "lower left",
    "lower right",
    "center left",
    "center right",
    "lower center",
    "upper center",
    "center",
]


class PlotSettings:
    """Old PlotSettings standin."""

    def migrate(self) -> dict:
        """Migrate class to dict."""
        dictionary = {}
        for key, value in self.__dict__.items():
            with contextlib.suppress(KeyError):
                key = PLOT_SETTINGS_MIGRATION_TABLE[key]
            if "scale" in key:
                value = 0 if value == "linear" else 1
            elif key == "legend_position":
                value = LEGEND_POSITIONS.index(value)
            dictionary[key] = value
        return dictionary


class ItemBase:
    """Old ItemBase standin."""

    def migrate(self) -> dict:
        """Migrate class to dict."""
        dictionary = {"type": self.item_type}
        for key, value in self.__dict__.items():
            with contextlib.suppress(KeyError):
                key = ITEM_MIGRATION_TABLE[key]
            if key in ITEM_VALUE_MIGRATION_TABLE:
                try:
                    value = ITEM_VALUE_MIGRATION_TABLE[key].index(value)
                except IndexError:
                    value = 0
            dictionary[key] = value
        return dictionary


class Item(ItemBase):
    """Old Item standin."""

    item_type = "GraphsDataItem"


class TextItem(ItemBase):
    """Old TextItem standin."""

    item_type = "GraphsTextItem"


_DEFAULT_VIEW = [0, 1, 0, 10, 0, 1, 0, 10]


def migrate_project(file: Gio.File) -> dict:
    """Migrate pickle-based project."""
    logging.debug("Migrating legacy project")
    sys.modules["graphs.misc"] = sys.modules[__name__]
    sys.modules["graphs.item"] = sys.modules[__name__]
    with gio_pyio.open(file, "rb") as wrapper:
        project = pickle.load(wrapper)

    figure_settings = project["plot_settings"].migrate()
    current_limits = [figure_settings[key] for key in LIMITS]
    history_pos = int(project["clipboard_pos"])
    history_states = _migrate_clipboard(
        project["datadict_clipboard"],
        history_pos,
        current_limits,
    )

    return {
        "version": str(project["version"]),
        "project-version": 1,
        "data": [item.migrate() for item in project["data"].values()],
        "figure-settings": figure_settings,
        "history-states": history_states,
        "history-position": history_pos,
        "view-history-states": [_DEFAULT_VIEW.copy(), current_limits],
        "view-history-position": -1,
    }


def _migrate_clipboard(clipboard, clipboard_pos, current_limits):
    if not clipboard:
        return []
    new_clipboard = []
    if len(clipboard) > 100:
        clipboard = list(clipboard[len(clipboard) - 100:])
    states = [{
        item.key: item.migrate()
        for item in state.values()
    } for state in clipboard]
    new_clipboard.append([[], _DEFAULT_VIEW.copy()])
    initial_items = states[1].values()
    new_clipboard.append((
        [[1, item] for item in initial_items],
        _get_limits(initial_items),
    ))
    if len(states) > 2:
        for count in range(len(states) - 2):
            batch = []
            previous_state = states[count + 1]
            current_state = states[count + 2]
            if len(current_state) < len(previous_state):
                for key, item in previous_state.copy().items():
                    if key not in current_state:
                        batch.append([2, [previous_state.index(item), item]])
                        previous_state.pop(item)
            for count_2, (key, item) in enumerate(current_state.items()):
                if key in previous_state:
                    previous_index = list(previous_state.keys()).index(key)
                    if previous_index != count_2:
                        batch.append([3, [previous_index, count_2]])
                    else:
                        for key_2, value in item.items():
                            previous_value = previous_state[key][key_2]
                            if value != previous_value:
                                batch.append([
                                    0,
                                    [key, key_2, previous_value, value],
                                ])
                else:
                    batch.append([1, item])
            if clipboard_pos == count - len(states) + 1:
                limits = _get_limits(current_state.values())
            else:
                limits = current_limits
            new_clipboard.append([batch, limits])
    return new_clipboard


def _get_limits(items):
    limits = [None] * 8
    for item in items:
        if item["type"] != "Item":
            continue
        for count, x_or_y in enumerate(["x", "y"]):
            index = item[f"{x_or_y}position"] * 2 + 4 * count
            data = item[f"{x_or_y}data"]
            try:
                limits[index] = min(limits[index], data)
            except TypeError:
                limits[index] = min(data)
            try:
                limits[index + 1] = max(limits[index + 1], data)
            except TypeError:
                limits[index + 1] = max(data)
    for count in range(8):
        default_view_copy = _DEFAULT_VIEW.copy()
        if limits[count] is None:
            limits[count] = default_view_copy[count]
    return limits



# SPDX-License-Identifier: GPL-3.0-or-later
"""Curve fitting module."""
from gettext import gettext as _

from gi.repository import Gio, Graphs

from graphs import canvas
from graphs.item import DataItem, FillItem

import numexpr

import numpy

from scipy.optimize import _minpack, curve_fit

import sympy

DATA_COLOR = "#1A5FB4"
FIT_COLOR = "#A51D2D"
FILL_COLOR = "#62A0EA"
FILL_ALPHA = 0.25
MARKER_STYLE = 1
MARKER_SIZE = 13
LINE_STYLE = 0


class CurveFittingDialog(Graphs.CurveFittingDialog):
    """Class for displaying the Curve Fitting dialog."""

    __gtype_name__ = "GraphsPythonCurveFittingDialog"

    def __init__(self, window: Graphs.Window, item: Graphs.Item):
        """Initialize the curve fitting dialog."""
        xdata, ydata = item.get_xydata()
        self._data = xdata, ydata
        x_min, x_max = min(xdata), max(xdata)
        padding = (x_max - x_min) * 0.025
        self._xlim = (x_min - padding, x_max + padding)
        self._x_fit = numpy.linspace(*self._xlim, 5000)

        style = Graphs.StyleManager.get_instance().get_system_style_params()
        self.data_curve = DataItem.new(
            style,
            xdata=xdata,
            ydata=ydata,
            name=item.get_name(),
            color=DATA_COLOR,
            linestyle=LINE_STYLE,
            markerstyle=MARKER_STYLE,
            markersize=MARKER_SIZE,
        )
        self.fitted_curve = DataItem.new(
            style,
            xdata=[],
            ydata=[],
            color=FIT_COLOR,
        )
        self.fill = FillItem.new(
            style,
            ([], [], []),
            color=FILL_COLOR,
            alpha=FILL_ALPHA,
        )
        self.residuals_item = DataItem.new(
            style,
            xdata=[],
            ydata=[],
            color=DATA_COLOR,
            linestyle=LINE_STYLE,
            markerstyle=MARKER_STYLE,
            markersize=MARKER_SIZE,
        )

        super().__init__(window=window)
        self.present(window)

    def _load_canvas(self) -> None:
        """Initialize and set main canvas."""
        settings = self.props.window.get_data().get_figure_settings()
        style = Graphs.StyleManager.get_instance().get_system_style_params()

        listmodel = Gio.ListStore.new(Graphs.Item)
        listmodel.append(self.fitted_curve)
        listmodel.append(self.fill)
        listmodel.append(self.data_curve)
        cv = canvas.Canvas(style, listmodel, interactive=False)
        ax = cv.figure.axis
        ax.set(
            xlabel=settings.get_bottom_label(),
            ylabel=settings.get_left_label(),
            xlim=self._xlim,
        )
        self.set_canvas(cv)

        listmodel = Gio.ListStore.new(Graphs.Item)
        listmodel.append(self.residuals_item)
        cv = canvas.Canvas(style, listmodel, interactive=False)
        ax = cv.figure.axis
        ax.set_ylabel(_("Residuals"))
        ax.set_xlabel(settings.get_bottom_label())
        ax.axhline(y=0, color="black", linestyle="--", linewidth=0.5)
        ax.set_xlim(*self._xlim)
        ax.set_ylim(-1, 1)
        cv.figure.props.legend = False
        self.set_residuals_canvas(cv)

    def _fit_curve(self) -> None:
        """Handle fit curve request."""
        free_vars = self.get_free_vars()
        variables = ["x"] + free_vars
        sym_params_map = dict(zip(variables, sympy.symbols(variables)))
        x_data, y_data = self._data
        equation = self.get_equation_string()
        settings = self.get_settings()

        def func(*params) -> numpy.ndarray:
            return numexpr.evaluate(equation, dict(zip(variables, params)))

        try:
            symbolic = sympy.sympify(equation, locals=sym_params_map)
            params, param_cov = curve_fit(
                func, x_data, y_data,
                p0=self.get_p0(),
                bounds=self.get_bounds(),
                nan_policy="omit",
                method=settings.get_string("optimization"),
            )
        except (sympy.SympifyError, TypeError, SyntaxError):
            self.set_results(Graphs.CurveFittingError.EQUATION)
            return
        except (RuntimeError, _minpack.error):
            self.set_results(Graphs.CurveFittingError.CONVERGENCE)
            return
        except (ValueError, ZeroDivisionError, OverflowError):
            self.set_results(Graphs.CurveFittingError.DOMAIN)
            return

        if numpy.any(numpy.isinf(param_cov)):
            self.set_results(Graphs.CurveFittingError.SINGULAR)
            return

        # Calculate statistics
        residuals = y_data - func(x_data, *params)
        ss_res = numpy.sum(residuals**2)
        ss_tot = numpy.sum((y_data - numpy.mean(y_data))**2)
        d_cov = numpy.sqrt(numpy.diagonal(param_cov))
        r2 = 1 - (ss_res / ss_tot)
        rmse = numpy.sqrt(ss_res / y_data.size)
        self.props.fit_result = Graphs.FitResult.new(params, d_cov, r2, rmse)
        self.residuals_item.set_xydata((x_data, residuals))

        # Substitute each free variables with the calculated value.
        values = dict(zip(free_vars, params))
        fitted_eq = str(sympy.simplify(symbolic.subs(values)))
        fitted_eq = Graphs.prettify_equation(fitted_eq)
        self.props.fitted_equation_string = fitted_eq

        x_fit = self._x_fit
        y_fit = func(x_fit, *params)
        if numpy.ndim(y_fit) == 0:
            y_fit = numpy.full(x_fit.size, y_fit.item())

        self.fitted_curve.set_xydata((x_fit, y_fit))
        self.fitted_curve.set_name(f"Y = {fitted_eq}")

        # Calculate and update confidence band for error propagation.
        local_dict = {"x": x_fit} | values
        jacobian = numpy.column_stack([
            numpy.full(x_fit.size, g) if numpy.ndim(
                g := numexpr.evaluate(
                    str(sympy.diff(symbolic, sym_params_map[name])),
                    local_dict,
                ),
            ) == 0 else g for name in free_vars
        ])
        variance = numpy.sum(jacobian * (jacobian @ param_cov), axis=1)

        std_dev_y = numpy.sqrt(numpy.abs(variance))
        confidence_band = std_dev_y * settings.get_enum("confidence")

        y_upper = y_fit + confidence_band
        y_lower = y_fit - confidence_band
        self.fill.set_data_tuple((x_fit, y_lower, y_upper))

        # Show fill and fit again after successful fit
        cv = self.get_canvas()
        ax = cv.figure.axis
        for line in ax.lines[1:]:
            line.set_visible(True)
        for collection in ax.collections:
            collection.set_visible(True)

        all_y = numpy.concatenate((y_lower, y_upper, y_data))
        all_y = all_y[numpy.isfinite(all_y)]
        y_min, y_max = all_y.min(), all_y.max()

        padding = (y_max - y_min) * 0.025
        ax.set_ylim(y_min - padding, y_max + padding)
        cv.queue_draw()

        cv = self.get_residuals_canvas()
        ax = cv.figure.axis
        ax.lines[0].set_visible(True)
        max_val = abs(residuals).max()
        if max_val > 0:
            y_lim = max_val * 1.1
            ax.set_ylim(-y_lim, y_lim)
        else:
            ax.set_ylim(-1, 1)
        cv.queue_draw()

        self.set_results(Graphs.CurveFittingError.NONE)

    def _clear_fit(self) -> None:
        """Clear all fit-related data by hiding curves."""
        cv = self.get_residuals_canvas()
        ax = cv.figure.axis
        ax.lines[0].set_visible(False)
        ax.set_ylim(-1, 1)

        # Hide all lines except the first one (data curve)
        cv = self.get_canvas()
        ax = cv.figure.axis
        for line in ax.lines[1:]:
            line.set_visible(False)
        # Hide all collections (fill)
        for collection in ax.collections:
            collection.set_visible(False)
        cv.queue_draw()



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for parsing files."""
import logging

from gi.repository import Graphs

_PARSERS = []


class Parser(Graphs.Parser):
    """Parser class."""

    __gtype_name__ = "GraphsPythonParser"

    def __init__(
        self,
        name: str,
        ui_name: str,
        filetype_name: str,
        file_suffixes: list[str],
    ):
        super().__init__(
            name=name,
            ui_name=ui_name,
            filetype_name=filetype_name,
            file_suffixes=file_suffixes,
        )

    @staticmethod
    def parse(_items, _settings, _data) -> None:
        """
        Parse a file given params.

        Must be implemented by parsers.
        """
        raise NotImplementedError

    @staticmethod
    def init_settings(_settings) -> bool:
        """Init settings."""
        return True

    @staticmethod
    def init_settings_widgets(_settings, _box) -> None:
        """Create settings widgets and append them to box."""


def register_parser(parser: Parser) -> None:
    """Register an import mode."""
    _PARSERS.append(parser)
    logging.debug("registered mode " + parser.get_name())


def list_parsers() -> list[Parser]:
    """List all registered parse modes."""
    return _PARSERS


def list_parser_names() -> list[str]:
    """List all localised parse names."""
    return [parser[0] for parser in _PARSERS]


def get_parser(index: int) -> Parser:
    """Get parser."""
    return _PARSERS[index]



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for parsing xrdml files."""
from gettext import gettext as _
from gettext import pgettext as C_

from gi.repository import Graphs

from graphs import file_io
from graphs.file_import.parsers import Parser
from graphs.item import DataItem

import numpy


class XrdmlParser(Parser):
    """Xrdml parser."""

    __gtype_name__ = "GraphsXrdmlParser"

    def __init__(self):
        super().__init__(
            "xrdml",
            C_("import-mode", "xrdml"),
            C_("file-filter", "PANalytical XRDML"),
            ["xrdml"],
        )

    @staticmethod
    def parse(
        items: Graphs.ItemList,
        settings: Graphs.ImportSettings,
        data: Graphs.Data,
    ) -> None:
        """Import data from xrdml file."""
        content = file_io.parse_xml(settings.get_file())
        intensities = content.getElementsByTagName("intensities")
        counting_time = content.getElementsByTagName("commonCountingTime")
        counting_time = float(counting_time[0].firstChild.data)
        ydata = intensities[0].firstChild.data.split()
        ydata = [int(value) / counting_time for value in ydata]

        scan_type = content.getElementsByTagName("scan")
        scan_axis = scan_type[0].attributes["scanAxis"].value
        if scan_axis.startswith("2Theta"):
            scan_axis = "2Theta"
        if scan_axis.startswith("Omega"):
            scan_axis = "Omega"

        data_points = content.getElementsByTagName("positions")
        for position in data_points:
            axis = position.attributes["axis"]
            if axis.value == scan_axis:
                unit = position.attributes["unit"].value
                start_pos = position.getElementsByTagName("startPosition")
                end_pos = position.getElementsByTagName("endPosition")
                start_pos = float(start_pos[0].firstChild.data)
                end_pos = float(end_pos[0].firstChild.data)
                xdata = numpy.linspace(start_pos, end_pos, len(ydata))
        items.add(
            DataItem.new(
                data.get_selected_style_params(),
                xdata,
                ydata,
                name=settings.get_filename(),
                xlabel=f"{scan_axis} ({unit})",
                ylabel=_("Intensity (cps)"),
            ),
        )
